#! /usr/bin/python
## -*- coding: utf-8 -*-

##
##  HARK Middleware
##  Copyright (C) 2013-2014 by Honda Research Institute Japan Co., Ltd.
##
##  $Id:$
##

import xml.dom.minidom
from .network import NetworkDef
from harkmw.defs.parameter import Parameters
from harkmw.HarkError import HARKError
from .helper import xml_attribute

class Document :
    """ HARK document definition """

    def __init__(self, typedefs) :
        self.typedefs   = typedefs
        self.parameters = Parameters()
        self.category   = ''
        self.comments   = ''
        self.networks   = {}
        self.mqtt_host  = 'localhost'
        self.mqtt_port  = 1883

    def load(self, nfile, xfile = None) :
        """ Load network definition from .n file
        """
        # Load into memory
        nfileString = ''
        with open(nfile) as f :
            isFirstLine = True
            for line in f :
                if not isFirstLine or "#!" not in line :
                    nfileString += line
                isFirstLine = False

        # Load into memory
        xfileString = ''
        if xfile is not None :
            with open(xfile) as f :
                isFirstLine = True
                for line in f :
                    if not isFirstLine or "#!" not in line :
                        xfileString += line
                    isFirstLine = False
        else :
            xfileString = nfileString

        # Parse XML 
        ndom = xml.dom.minidom.parseString(nfileString)
        xdom = xml.dom.minidom.parseString(xfileString)
        self.loadXML(ndom.documentElement, xdom.documentElement)

    def get_network_def(self, name) :
        return self.networks[name] if name in self.networks else None

    def loadXML(self, nroot, xroot) :
        """ Load network definition from XML """

        # Try to fetch 'category' and 'comments'
        if nroot.getAttributeNode('category') :
            self.category = nroot.attributes['category']

        if nroot.getAttributeNode('comments') :
            self.comments = nroot.attributes['comments']

        # Process each 'depth=1' elements
        for child in nroot.childNodes :
            if child.nodeType != child.ELEMENT_NODE :
                continue

            if child.nodeName == 'Network' :
                network = NetworkDef(self)
                network.loadXML(child)
                self.networks[network.name] = network
            
            elif child.nodeName == 'IncludeNetwork' :
                raise NotImplementedError()
            
            elif child.nodeName == 'Parameter' :
                if not child.getAttributeNode('name') :
                    raise HARKError("Node <%s> has no attribute '%s'" % (child.nodeName, 'name'))
                else :
                    parameterName = child.attributes['name'].nodeValue

                if not child.getAttributeNode('type') :
                    raise HARKError("Node <%s> has no attribute '%s'" % (child.nodeName, 'value'))
                else :
                    parameterType = child.attributes['type'].nodeValue

                if not child.getAttributeNode('value') :
                    raise HARKError("Node <%s> has no attribute '%s'" % (child.nodeName, 'value'))
                else :
                    parameterValue = child.attributes['value'].nodeValue

                self.parameters.add(parameterName, parameterType, parameterValue)
                
            else :
                pass # .x ファイルを読み込んでいるケースもあるので、エラーにするのではなく無視する
                #raise HARKError("Unexpected node <%s> in definition" % child.nodeName)

        for child in xroot.childNodes :
            if child.nodeType != child.ELEMENT_NODE :
                continue

            if child.nodeName == 'MQTT' :
                for ch in child.childNodes :
                    if ch.nodeType != child.ELEMENT_NODE :
                        continue

                    elif ch.nodeName == 'Broker' :
                        self.mqtt_host = xml_attribute(ch, 'host', False, self.mqtt_host)
                        self.mqtt_port = int(xml_attribute(ch, 'port', False, self.mqtt_port))

            elif child.nodeName == 'Network' :
                name = xml_attribute(child, 'name')
                if name not in self.networks :
                    raise HARKError("Document does not contain network with name '%s'" % name)

                else :
                    network = self.networks[name]
                    network.process = xml_attribute(child, 'process', False, None)
#                   print network.nodes

                    for ch in child.childNodes :
                        if ch.nodeType != child.ELEMENT_NODE :
                            continue

                        elif ch.nodeName == 'Node' :
                            name = xml_attribute(ch, 'name')
                            if name not in network.nodes :
                                raise HARKError("Network does not contain node with name '%s'" % name)

                            else :
                                node = network.nodes[name]
                                node.process = xml_attribute(ch, 'process', False, None)
                                node.topic   = xml_attribute(ch, 'topic', False, None)

    def getNetTerminals(self, input, type) :
        return []


    def dump(self, prefix='') :
        print(prefix + "[Document]")
        print(prefix + "  category=%s" % self.category)
        print(prefix + "  comments=%s" % self.comments)
        print(prefix + "  mqtt_host=%s" % self.mqtt_host)
        print(prefix + "  mqtt_port=%d" % self.mqtt_port)
        print(prefix + "  [Parameters]")
        self.parameters.dump(prefix + "  ")
        for network in self.networks.values() :
            network.dump(prefix + "  ")

# end of file
