#! /usr/bin/python
## -*- coding: utf-8 -*-

##
##  HARK Middleware
##  Copyright (C) 2013-2014 by Honda Research Institute Japan Co., Ltd.
##
##  $Id:$
##

from harkmw.HarkError      import HARKError
from harkmw.defs.parameter import Parameters
from harkmw.logic.network  import Network
from harkmw.logic.proxy    import Proxy
from harkmw.logic.proxy    import ProxyTerminal
from harkmw.logic.link      import Link
from harkmw.logic.mqtt_node import MQTTNode
from harkmw.types.native_node import NativeNode
#from harkmw.logic.terminal import NetTerminal

class Builder :
    def __init__(self, process) :
        self.process = process

    def build(self, document, params, process_name) :
        # retrieve network named "MAIN"
        parameters = params.clone().update(document.parameters)
#       parameters.dump()

        if "MAIN" not in document.networks :
            raise HARKError("Network with name 'MAIN' not defined")
        else :
            network_def = document.networks["MAIN"]

        # build main network
        network = self.build_network(document, None, network_def, parameters, process_name)
#        network.dump()

        return network

    def build_network(self, document, parent_network, network_def, params, process_name) :
        """ Build network object from network definition
        """
        #print "Build network with name '%s'" % network_def.name
        #params.dump()
        network = Network()
        network.parent_network = parent_network
        network.network_type = network_def.type
        network.name = network_def.name
        network.topic = network_def.topic
        network.process_name = network_def.get_process_name()

        if not network_def.outputs:
            raise HARKError("No output terminal in network '%s'" % network_def.name)
        if network_def.type == 'iterator' and not network_def.conditions:
            raise HARKError("No condition terminal in iterator network '%s'" % network_def.name)

        # build nodes
        for node_def in network_def.nodes.values() :
            node = self.build_node(document, network, node_def, params, process_name)
            #print "Append node with name %s" % node.name
            if node is not None :
                network.nodes[node.name] = node

        # build terminals
        output_collector = NativeNode()
        input_collector  = NativeNode()
        if network_def.type == 'iterator':
            output_collector.build_type(network, "HARKMW_Collector", "%s_output" % network.name)
            input_collector. build_type(network, "HARKMW_Collector", "%s_input"  % network.name)
        else:
            output_collector.build_type(network, "Collector", "%s_output" % network.name)
            input_collector. build_type(network, "Collector", "%s_input"  % network.name)
        network.output_collector = output_collector
        network.input_collector  = input_collector
        network.nodes["ALL_NETWORK_OUTPUT"] = output_collector
        network.nodes["ALL_NETWORK_INPUT"]  = input_collector

        if network_def.type == 'iterator' :
            condition_collector = NativeNode()
            condition_collector.build_type(network, "Collector", "%s_condition" % network.name)
            network.condition_collector = condition_collector
            network.nodes["NETWORK_CONDITION"] = condition_collector

#       self.inputs  = {}
#       self.outputs = {}
        for (externalName, input) in network_def.inputs.items() :
            node         = network.nodes[input.node.name]
            #nodeTerminal = node.inputs  [input.name]
#           self.connect_terminal(network, network, externalName, node, input.name)
            self.connect_link(network, input_collector, externalName, node, input.name)

        for (externalName, output) in network_def.outputs.items() :
            node         = network.nodes[output.node.name]
            #nodeTerminal = node.outputs [output.name]
#           self.connect_terminal(network, node, output.name, network, externalName)
            self.connect_link(network, node, output.name, output_collector, externalName)

        self.conditions = {}
        for (externalName, condition) in network_def.conditions.items() :
            node         = network.nodes[condition.node.name]
            #netTerminal = NetTerminal(network.nodes[condition.node.name], condition.name, None)
            #self.conditions[externalName] = condition
            #print "CONDITION: %s" % externalName
            self.connect_link(network, node, condition.name, condition_collector, "OUTPUT")

        # build links
        for link_def in network_def.links :
            link = self.build_link(document, network, link_def, process_name)
            if link is not None :
                network.links.append(link)

#       print "Build network with name '%s' completed" % network_def.name
#       print "Output: %s" % network.outputs
        return network

    def build_node(self, document, network, node_def, params, process_name) :
        """ Build node object from node definition
        """
        # build node
        node = None

        #print node_def.get_process_name()
        #print "%s: %s" % (node_def.name, node_def.get_process_name())
        node_def.parameters.normalize(params)
        if node_def.get_process_name() == process_name :
            # is C++-implemented node?
            native_node = NativeNode()
            try:
                native_node.build(network, node_def, params)
                return native_node
            except NotImplementedError :
                # is Python-implemneted node?
                pass
        
        # is network?
        network_def = node_def.network.document.get_network_def(node_def.type)
        if network_def is None and node_def.type in document.typedefs.nodedefs :
            network_def = document.typedefs.nodedefs[node_def.type]

        if network_def is not None :
            parameters = node_def.parameters.clone().update(params)
            node = self.build_network(document, network, network_def, parameters, process_name)
            node.name = node_def.name # override network's name with node's
            node.type = node_def.type # override network's type with node's
            return node

        # is MQTT node ?
        if node_def.get_process_name() != process_name :
            node = MQTTNode()
            node.build(network, node_def, params)
            return node
        
        raise HARKError("No node/network definition with name '%s' found" % node_def.type)

    def build_link(self, document, network, link_def, process_name) :
        """ Build link object from link definition
        """
        # retrieve src node
        if link_def.pinFrom.node.name in network.nodes :
            nodeFrom = network.nodes[link_def.pinFrom.node.name]
#             pinFrom  = nodeFrom.outputs[link_def.pinFrom.name]
        else :
#           nodeFrom = None
            raise HARKError("Node with name '%s' not defined" % link_def.pinFrom.node.name)

        # retrieve dst node
        if link_def.pinTo.node.name in network.nodes :
            nodeTo = network.nodes[link_def.pinTo.node.name]
#             pinTo  = nodeTo.inputs[link_def.pinTo.name]
        else :
#            nodeTo = None
            raise HARKError("Node with name '%s' not defined" % link_def.pinTo.node.name)

#         # build link
#         terminalFrom = nodeFrom.outputs[link_def.pinFrom.name]
#         terminalTo   = nodeTo.inputs[link_def.pinTo.name]

        return self.connect_link(network, nodeFrom, link_def.pinFrom.name, nodeTo, link_def.pinTo.name)

    def connect_link(self, network, node_from, terminal_name_from, node_to, terminal_name_to) :
        """
        """
        proxy = Proxy(self.process)
#       print "connect_link: from=%s/%s/%s, to=%s/%s/%s" % (node_from.type, node_from.name, terminal_name_from, node_to.type, node_to.name, terminal_name_to)

        if node_from is None :
            pass
        elif isinstance(node_from, Network) :
            node_from = node_from.output_collector

        if node_from is None :
            pass

        elif isinstance(node_from, NativeNode) :
            proxy.connect_native_input(node_from, terminal_name_from)

        elif isinstance(node_from, Network) :
            #proxy.connect_network_input(node_from, terminal_name_from)
            proxy.connect_native_input(node_from.output_collector, terminal_name_from)

        else :
            proxy.connect_node_input(node_from, terminal_name_from)

        if node_to is None :
            pass
        elif isinstance(node_to, Network) :
            node_to.inputs[node_to.input_collector.name] = proxy
            node_to = node_to.input_collector

        if node_to is None :
            pass

        elif isinstance(node_to, NativeNode) :
            proxy.connect_native_output(node_to, terminal_name_to)

        elif isinstance(node_to, Network) :
            #proxy.connect_network_output(node_to, terminal_name_to)
            proxy.connect_native_output(node_to.input_collector, terminal_name_to)

        else :
            proxy.connect_node_output(node_to, terminal_name_to)

        if node_from is not None :
            if terminal_name_from not in node_from.outputs :
                node_from.outputs[terminal_name_from] = [proxy]
            else :
                node_from.outputs[terminal_name_from].append(proxy)

        if node_to is not None :
            node_to.inputs[terminal_name_to] = proxy

    def connect_terminal(self, network, node_from, terminal_name_from, node_to, terminal_name_to) :
        """
        """
        from harkmw.typelib.native_node import NativeNode

#       print "connect_terminal: from=%s, to=%s" % (("%s/%s/%s" % (node_from.type, node_from.name, terminal_name_from) if node_from is not None else "(none)"), ("%s/%s/%s" % (node_to.type, node_to.name, terminal_name_to) if node_to is not None else "(none)"))
        if node_from is not None :
            if isinstance(node_from, Network) :
#               print "%s/%s is a network" % (node_from.type, node_from.name)
                node_from.input_terminals[terminal_name_from] = ProxyTerminal(node_to, terminal_name_to)

            else :
#               print "%s/%s is not a network" % (node_from.type, node_from.name)
                pass
        
        if node_to is not None :
            if isinstance(node_to, Network) :
#               print "%s/%s is a network" % (node_to.type, node_to.name)
                node_to.output_terminals[terminal_name_to] = ProxyTerminal(node_from, terminal_name_from)

            else :
#               print "%s/%s is not a network" % (node_to.type, node_to.name)
                pass

# end of file
