#! /usr/bin/python
## -*- coding: utf-8 -*-

##
##  HARK Middleware
##  Copyright (C) 2013-2014 by Honda Research Institute Japan Co., Ltd.
##
##  $Id:$
##

import pickle
import sys

from harkmw.defs.document  import Document
from harkmw.defs.type      import Typedefs
from harkmw.defs.parameter import Parameters
from harkmw.logic.builder  import Builder
from harkmw.logic.network  import Network
from harkmw.logic.proxy    import Proxy
from harkmw.types.dynamic  import Dynamic

class Process :
    def __init__(self, args) :
        self.typedefs = Typedefs()

        params = Parameters()
        for i, argv in enumerate(args) :
            params.add('ARG%d'        % (i + 1), 'string', argv)
            params.add('string:ARG%d' % (i + 1), 'string', argv)
            params.add('int:ARG%d'    % (i + 1), 'int',    argv) # ToDo (in C++, = Int::alloc(atoi(argv[arg]))) 
            params.add('float:ARG%d'  % (i + 1), 'float',  argv) # ToDo (in C++, = Float::alloc(atof(argv[arg]))) 
            params.add('object:ARG%d' % (i + 1), 'object', argv) # ToDo (in C++, see batchflow.cc:75)
#       params.dump()
        self.params = params
        self.toolbox_dirs = []
        self.mosquitto_client = None
        self.mqtt_topics = {}
        self.mqtt_queue  = []

    def add_toolbox_dirs(self, dirs) :
        self.toolbox_dirs.extend(dirs)

    def load(self, nfile, xfile) :
        self.nfile = nfile
        self.xfile = xfile

        # Scan toolboxes
        for dir in self.toolbox_dirs :
#           print dir
            self.scan_nodedefs(dir)
            self.scan_typedefs(dir)
            self.scan_typelibs(dir)

        # Load .n/.x file
        document = Document(self.typedefs)
        document.load(self.nfile, self.xfile)
#       document.dump()
        self.document = document

    def run(self, process_name) :
        self.process_name = process_name

        # Build network
        builder = Builder(self)
        network = builder.build(self.document, self.params, process_name)

        try:
            # Run network
            network.request(self)
            network.initialize(self)
            network.run(self)
            network.release()
#           print "Network released"

        finally:
#           print "Set running flag to False"
            self.running = False

        if self.is_mqtt_initialized() :
            self.finalize_mqtt()

    def initialize_mqtt(self) :
        # Setup MQTT
        import paho.mqtt.client as mqtt
        self.mosquitto_subscribers = {}
        self.mosquitto_client = mqtt.Client('HARK Middleware Client: %s' % self.process_name)
        self.mosquitto_client.user_data_set(self)
        self.mosquitto_client.on_message = self.on_mqtt_message
        self.mosquitto_client.connect(self.document.mqtt_host, self.document.mqtt_port)

    def finalize_mqtt(self) :
        # Close MQTT
        while self.mosquitto_client.want_write() :
            self.mosquitto_client.loop_write()
        self.mosquitto_client.loop()
        self.mosquitto_client.disconnect()
        self.mosquitto_client = None

    def is_mqtt_initialized(self) :
        return self.mosquitto_client is not None

    def on_mqtt_message(self, mosquitto, obj, msg) :
#       print "Message received on topic "+msg.topic+" with QoS "+str(msg.qos)+" and payload "+msg.payload
#        print "Message received on topic "+msg.topic
        self.mqtt_queue.append(msg)

    def process_mqtt_message(self, msg) :
        payload = msg.payload
        marker  = payload[0:4]
        version = payload[4:8]
        type    = payload[8:12]
        data    = payload[12:]

#       print "on_mqtt_message: data=%s" % data
        obj = pickle.loads(data)
        if not isinstance(obj, dict) :
            raise HARKError("Received non-dictionary data via MQTT")

#       print "obj: %s" % obj
        topic = obj['topic']
        count = obj['count']
        body  = obj['data']

#       print "Receive: %s, %d, %s" % (topic, count, body)
        print("Receive: %s, %d" % (topic, count))

#       print "Topic: %s, %s" % (topic, (topic in self.mosquitto_subscribers))
        if topic in self.mosquitto_subscribers :
            for receiver, args in self.mosquitto_subscribers[topic] :
#               print "Send to: %s, %s" % (s['proxy'], s)
                if isinstance(receiver, Proxy) :
                    receiver.input_node.network.add_remote_data(count, receiver, args, body);
                elif isinstance(receiver, Network) :
#                   self.mqtt_topics[topic] = data
#                   receiver.process_mqtt_message(topic, count, body, args, self)
                    receiver.add_mqtt_message(topic, count, body, args, self)

#     def thread_main(self) :
#         while self.running :
# #           print "Checking loop"
#             self.mosquitto_client.loop(100)
# #           print "Checking loop ok: %s" % self.running

    def check_mqtt(self) :
        self.mosquitto_client.loop(10)
        while len(self.mqtt_queue) > 0 :
            self.process_mqtt_message(self.mqtt_queue.pop())

    def wait_topic(self, topic) :
        if topic not in self.mqtt_topics :
#           print "Waiting: %s" % topic
            while topic not in self.mqtt_topics :
                self.check_mqtt()

        return self.mqtt_topics.pop(topic)

    def publish(self, topic, count, data) :
#       print "Publish: %s, %d, %s" % (topic, count, data)
        print("Publish: %s, %d" % (topic, count))
        p = 'HARK' + '0001' + 'PCKL' + pickle.dumps({'count': count, 'data': data, 'topic': topic})
        result, qid = self.mosquitto_client.publish(topic, p, 1)
#        print "Result: %d, ID: %d" % (result, qid)

    def subscribe(self, topic, receiver, args=None) :
        if not self.is_mqtt_initialized() :
            self.initialize_mqtt()

        if topic not in self.mosquitto_subscribers :
            self.mosquitto_subscribers[topic] = []

        self.mosquitto_subscribers[topic].append((receiver, args))
        self.mosquitto_client.subscribe(topic, 0)

        if isinstance(receiver, Proxy) :
            print("Subscribe: %s -> %s/%s" % (topic, receiver.output_node.name, receiver.output_terminal_name))
        else :
            print("Subscribe: %s -> %s" % (topic, receiver))

    def scan_nodedefs(self, path) :
        import os
        for (root, dirs, files) in os.walk(path, True, None, True) :
            for file in files:
                if file[-2:] == ".n":
#                   print "Loading .n file %s" % file
                    self.typedefs.load_nodedef(os.path.join(root, file))

    def scan_typelibs(self, path) :
        import os
        for (root, dirs, files) in os.walk(path, True, None, True) :
            for file in files:
                if "tlb" in file.split("."):
#                   print "Loading .tlb file %s" % file
                    Dynamic().load_dynamic(os.path.join(root, file))

    def scan_typedefs(self, path) :
        import os
        for (root, dirs, files) in os.walk(path, True, None, True) :
            for file in files:
                if file[-4:] == ".def":
#                   print "Loading .def file %s" % file
                    self.typedefs.load_typedef(os.path.join(root, file))
#       self.typedefs.dump()

    def print_modules(self, dirs) :
        for path in dirs :
            self.scan_typelibs(path)
        Dynamic().print_modules()

# end of file
